class Selection:
    """
    Class implementing the selection operator for the genetic algorithm.

    Args:
        population (list): List of trees.
        selection_tournament_size (int): Number of trees to select.
        random (Random): Random number generator.
    """
    def selection(population, selection_tournament_size, random):
        """
        Selection is the process of choosing parent trees from the current population to produce offspring for the next generation. By default, `GATree` class uses tournament selection, a method where a subset of the population is randomly chosen, and the best individual from this subset is selected.

        Pseudocode of the implementation:

        1. Randomly select `selection_tournament_size` trees from the population.
        2. Evaluate the fitness of the selected trees.
        3. Choose the tree with the best fitness from the selected subset.
        4. Repeat the process to select another parent.
        5. Ensure the selected parents are different to maintain genetic diversity.

        Args:
            population (list): List of trees.
            selection_tournament_size (int): Number of trees to select.
            random (Random): Random number generator.

        Returns:
            tuple: The two selected trees.
        """
        valid = False
        while not valid:
            selection = []

            # Select two trees
            for _ in range(2):
                indices = random.choice(len(population),
                                        selection_tournament_size, replace=False)
                candidates = [population[i] for i in indices]
                candidates.sort(key=lambda x: x.fitness, reverse=False)
                selection.append(candidates[0])

            # Check if trees are different
            if selection[0] != selection[1]:
                valid = True

        return selection[0], selection[1]



import numpy as np
from sklearn.base import BaseEstimator

# GATree
from gatree.tree.node import Node


class GATree(BaseEstimator):
    """
    Evolutionary decision tree classifier. The GATree classifier is a decision tree classifier that is trained using a genetic algorithm. The genetic algorithm is used to evolve a population of trees over multiple generations. The fitness of each tree is evaluated using a fitness function, which is used to select the best trees for crossover and mutation.

    Args:
        max_depth (int, optional): Maximum depth of the tree.
        random (Random, optional): Random number generator.
        fitness_function (function, optional): Fitness function for the genetic algorithm.
        n_jobs (int, optional): Number of jobs to run in parallel.
        random_state (int, optional): Seed for reproducibility.

    Attributes:
        max_depth (int, optional): Maximum depth of the tree.
        random (Random): Random number generator.
        X (pandas.DataFrame): Training data.
        y (pandas.Series): Target values.
        att_indexes (numpy.ndarray): Array of attribute indexes.
        att_values (dict): Dictionary of attribute values.
        class_count (int): Number of classes.
        fitness_function (function): Fitness function for the genetic algorithm.
        n_jobs (int): Number of jobs to run in parallel.
        random_state (int): Seed for reproducibility.
        _tree (Node): The fitted tree.
        _best_fitness (list): List of best fitness values for each iteration.
        _avg_fitness (list): List of average fitness values for each iteration.
    """

    def __init__(self, max_depth=None, random=None, fitness_function=None, n_jobs=1, random_state=None):
        """
        Initialise the Genetic Algorithm Tree Classifier. Maximum depth, random number generator, fitness function, number of jobs, and random state can be specified.

        Args:
            max_depth (int, optional): Maximum depth of the tree.
            random (Random, optional): Random number generator.
            fitness_function (function, optional): Fitness function for the genetic algorithm.
            random_state (int, optional): Seed reproducibility.
        """
        self.max_depth = max_depth
        if random is None and random_state is not None:
            np.random.seed(random_state)
        self.random = random if random is not None else np.random
        self.fitness_function = fitness_function if fitness_function is not None else self.default_fitness_function
        self.n_jobs = n_jobs
        self._tree = None
        self._best_fitness = []
        self._avg_fitness = []

    @staticmethod
    def default_fitness_function(root, **fitness_function_kwargs):
        """ 
        Default fitness function for the genetic algorithm.

        Args:
            root (Node): Root node of the tree.

        Returns:
            float: The fitness value.
        """
        pass

    @staticmethod
    def _predict_and_evaluate(tree, X, y, fitness_function, is_training=False, **fitness_function_kwargs):
        """
        Evaluate a tree on a training set (in parallel).

        Args:
            tree (Node): Tree to evaluate.
            X (pandas.DataFrame): Training data.
            y (pandas.Series): Target values.
            fitness_function (function): Fitness function for the genetic algorithm.
            is_training (bool): If the instances are used for training or predicting.

        Returns:
            Node: The evaluated tree.
        """
        for j in range(X.shape[0]):
            # Predict class for current instance
            tree.predict_one(X.iloc[j], y.iloc[j], is_training)
        tree.fitness = fitness_function(tree, **fitness_function_kwargs)
        return tree

    def fit(self, X, y, population_size=150, max_iter=2000, mutation_probability=0.1, elite_size=1,
            selection_tournament_size=2, fitness_function_kwargs={}):
        """
        Fit a tree to a training set. The population size, maximum iterations, mutation probability, elite size, and selection tournament size can be specified.

        Args:
            X (pandas.DataFrame): Training data.
            y (pandas.Series): Target values.
            population_size (int, optional): Size of the population.
            max_iter (int, optional): Maximum number of iterations.
            mutation_probability (float, optional): Probability of mutation.
            elite_size (int, optional): Number of elite trees.
            selection_tournament_size (int, optional): Number of trees in tournament.
            fitness_function_kwargs (dict, optional): Additional kwargs to be passed to the fitness_funciton.

        Returns:
            Node: The fitted tree.
        """
        pass

    def predict(self, X):
        """
        Predict classes for the given data.

        Args:
            X (pandas.DataFrame): Data to predict.

        Returns:
            list: Predicted classes.
        """
        y_pred = []
        for i in range(X.shape[0]):
            index = self._tree.predict_one(X.iloc[i])
            y_pred.append(self.att_values[-1][index])
        return y_pred

    def plot(self, node=None, prefix=''):
        """
        Plot the decision tree with nodes and leaves.

        Args:
            node (Node, optional): Current node to plot.
            prefix (str, optional): Prefix for the current node.
        """
        if node is None:
            node = self._tree

        if node is not None:
            if node.att_index != -1:
                print(prefix + '├── {} > {}'.format(self.X.columns.tolist()
                      [node.att_index], node.att_value))
            else:
                print(
                    prefix + '└── Class: {}'.format(self.att_values[-1][node.att_value]))

            if node.left is not None or node.right is not None:
                self.plot(node.left, prefix + '    ')
                self.plot(node.right, prefix + '    ')



import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.metrics import silhouette_score
from sklearn.base import ClusterMixin

# GATree
from gatree.tree.node import Node
from gatree.ga.selection import Selection
from gatree.ga.crossover import Crossover
from gatree.ga.mutation import Mutation
from gatree.gatree import GATree


class GATreeClustering(ClusterMixin, GATree):
    """
    Evolutionary decision tree clustering. The GATree clustering is a decision tree clustering that is trained using a
    genetic algorithm. The genetic algorithm is used to evolve a population of trees over multiple generations.
    The fitness of each tree is evaluated using a fitness function, which is used to select the best trees for
    crossover and mutation.

    Args:
        max_depth (int, optional): Maximum depth of the tree.
        random (Random, optional): Random number generator.
        fitness_function (function, optional): Fitness function for the genetic algorithm.
        n_jobs (int, optional): Number of jobs to run in parallel.
        random_state (int, optional): Seed for reproducibility.
        min_clusters (int, optional): Number of minimum clusters.
        max_clusters (int, optional): Number of maximum clusters.

    Attributes:
        max_depth (int, optional): Maximum depth of the tree.
        random (Random): Random number generator.
        X (pandas.DataFrame): Training data.
        y (pandas.Series): Target values.
        att_indexes (numpy.ndarray): Array of attribute indexes.
        att_values (dict): Dictionary of attribute values.
        class_count (int): Number of classes.
        fitness_function (function): Fitness function for the genetic algorithm.
        n_jobs (int): Number of jobs to run in parallel.
        random_state (int): Seed for reproducibility.
        min_clusters (int): Number of minimum clusters.
        max_clusters (int): Number of maximum clusters.
        _tree (Node): The fitted tree.
        _best_fitness (list): List of best fitness values for each iteration.
        _avg_fitness (list): List of average fitness values for each iteration.
    """

    def __init__(self, max_depth=None, random=None, fitness_function=None, n_jobs=1, random_state=None, min_clusters=2,
                 max_clusters=10):
        """
        Initialise the Genetic Algorithm Tree Clustering. Maximum depth, random number generator, fitness function, number of jobs, and random state can be specified.

        Args:
            max_depth (int, optional): Maximum depth of the tree.
            random (Random, optional): Random number generator.
            fitness_function (function, optional): Fitness function for the genetic algorithm.
            random_state (int, optional): Seed reproducibility.
            min_clusters (int, optional): Minimum number of clusters.
            max_clusters (int, optional): Maximum number of clusters.
        """
        super().__init__(max_depth, random, fitness_function, n_jobs, random_state)
        self.min_clusters = min_clusters
        self.max_clusters = max_clusters

    @staticmethod
    def default_fitness_function(root, **fitness_function_kwargs):
        """
        Default fitness function for the genetic algorithm.

        Args:
            root (Node): Root node of the tree.

        Returns:
            float: The fitness value.
        """
        if len(set(root.y_pred)) < fitness_function_kwargs['min_clusters']:
            return 1

        return 1 - ((silhouette_score(fitness_function_kwargs['fitness_X'], root.y_pred) + 1) / 2) + (0.002 * root.size())

    def fit(self, X, population_size=150, max_iter=2000, mutation_probability=0.1, elite_size=1,
            selection_tournament_size=2, fitness_function_kwargs={}):
        """
        Fit a tree to a training set. The population size, maximum iterations, mutation probability, elite size, and selection tournament size can be specified.

        Args:
            X (pandas.DataFrame): Training data.
            y (pandas.Series): Target values.
            population_size (int, optional): Size of the population.
            max_iter (int, optional): Maximum number of iterations.
            mutation_probability (float, optional): Probability of mutation.
            elite_size (int, optional): Number of elite trees.
            selection_tournament_size (int, optional): Number of trees in tournament.
            fitness_function_kwargs (dict, optional): Additional kwargs to be passed to the fitness_funciton.

        Returns:
            Node: The fitted tree.
        """
        self.X = X
        none_series = pd.Series([None] * len(X))
        self.y = none_series
        self.att_indexes = np.arange(X.shape[1])
        self.att_values = {i: [(min_val + max_val) / 2 for min_val, max_val in zip(sorted(
            X.iloc[:, i].unique())[:-1], sorted(X.iloc[:, i].unique())[1:])] for i in range(X.shape[1])}
        self.att_values[-1] = list(range(self.max_clusters))
        self.class_count = self.max_clusters

        # Generation of initial population
        node = Node()
        population = []
        for _ in range(population_size):
            population.append(node.make_node(max_depth=self.max_depth, random=self.random,
                              att_indexes=self.att_indexes, att_values=self.att_values, class_count=self.class_count))

        for i in range(max_iter+1):
            # Clear previous evaluation
            for tree in population:
                tree.clear_evaluation()

            # Evaluation of population
            population = Parallel(n_jobs=self.n_jobs)(delayed(GATreeClustering._predict_and_evaluate)(
                tree, X, none_series, self.fitness_function, True, **fitness_function_kwargs) for tree in population)

            # Sort population by fitness
            population.sort(key=lambda x: x.fitness, reverse=False)

            # Log best and average fitness
            self._best_fitness.append(population[0].fitness)
            self._avg_fitness.append(
                sum([tree.fitness for tree in population]) / len(population))

            if i != max_iter:
                # Elites
                elites = population[:elite_size]

                # Descendant generation
                descendant = []
                for _ in range(0, len(population), 2):
                    # Tournament selection
                    tree1, tree2 = Selection.selection(
                        population=population, selection_tournament_size=selection_tournament_size, random=self.random)

                    # Crossover between selected trees
                    crossover1 = Crossover.crossover(
                        tree1=tree1, tree2=tree2, random=self.random)
                    crossover2 = Crossover.crossover(
                        tree1=tree2, tree2=tree1, random=self.random)

                    # Mutation of new trees
                    mutation1 = crossover1
                    mutation2 = crossover2
                    if self.random.random() < mutation_probability:
                        mutation1 = Mutation.mutation(root=crossover1, att_indexes=self.att_indexes,
                                                      att_values=self.att_values, class_count=self.class_count,
                                                      random=self.random)
                    if self.random.random() < mutation_probability:
                        mutation2 = Mutation.mutation(root=crossover2, att_indexes=self.att_indexes,
                                                      att_values=self.att_values, class_count=self.class_count,
                                                      random=self.random)

                    # Add new trees to descendant population
                    descendant.extend([mutation1, mutation2])

                # Elites + descendants
                descendant.sort(key=lambda x: x.fitness, reverse=False)
                descendant = elites + descendant[:population_size - elite_size]

                # Replace old population with new population
                population = descendant

        self._tree = population[0]



from gatree.tree.node import Node


class Crossover:
    """
    Class implementing the crossover operator for the genetic algorithm.
    """
    def crossover(tree1, tree2, random):
        """
        Crossover is a genetic operator used to combine the genetic information of two parent trees to generate new offspring. This enables exploration, which helps in creating diversity in the population and combining good traits from both parents.

        Pseudocode of the implementation:

        1. Copy the parent trees to avoid altering the originals.
        2. Randomly select a crossover point in each tree.
        3. Swap the subtrees at the selected points between the two trees.
        4. Return the new tree created from the crossover.

        Args:
            tree1 (Node): The first tree for crossover.
            tree2 (Node): The second tree for crossover.
            random (Random): Random number generator.

        Returns:
            Node: The new tree resulting from crossover.
        """
        n1 = Node.copy(tree1)
        n2 = Node.copy(tree2)
        size1 = n1.max_depth()
        size2 = n2.max_depth()

        while True:
            if (n1.left == None or random.randint(0, size1) == 0) and n1.parent != None:
                break

            if random.choice([True, False]):
                n1 = n1.left
            else:
                n1 = n1.right

        while True:
            if (n2.left == None or random.randint(0, size2) == 0) and n2.parent != None:
                break

            if random.choice([True, False]):
                n2 = n2.left
            else:
                n2 = n2.right

        p = n1.parent
        if p.left == n1:
            p.set_left(n2)
        else:
            p.set_right(n2)

        return p.get_root()



class Node:
    """
    Node class for decision trees. Each node represents a decision in the tree. The node can be a mid-tree node or a leaf node. Mid-tree nodes have an attribute index and an attribute value, while leaf nodes have a class index. The node can have a parent, left child, and right child. The node also has a fitness value, true class values, and predicted class values for evaluation.

    Args:
        att_index (int, optional): Attribute index or -1 if this is a leaf.
        att_value (any, optional): Attribute value or class index if this is a leaf.

    Attributes:
        parent (Node): Parent node.
        left (Node): Left child node.
        right (Node): Right child node.
        att_index (int, optional): Attribute index or -1 if this is a leaf.
        att_value (any, optional): Attribute value or class index if this is a leaf.
        fitness (None): Placeholder for fitness value.
        y_true (list): List of true class values for evaluation.
        y_pred (list): List of predicted class values for evaluation.
    """

    def __init__(self, att_index=None, att_value=None):
        """
        Initializes the node with the given parameters. If the attribute index is -1, then this is a leaf node.

        Args:
            att_index (int, optional): Attribute index or -1 if this is a leaf.
            att_value (any, optional): Attribute value or class index if this is a leaf.
        """
        if att_index is None and att_value is not None:
            self.__init__(-1, att_value)  # leaf node
        else:
            self.parent = None
            self.left = None
            self.right = None

            self.att_index = att_index if att_index is not None else None  # if -1 then this is leaf
            self.att_value = att_value if att_value is not None else None

            self.fitness = None
            self.y_true = []
            self.y_pred = []

    @staticmethod
    def copy(node, parent=None):
        """
        Returns a deep copy of the given node. The parent node can be specified for the copied node.

        Args:
            node (Node): Node to be copied.
            parent (Node, optional): Parent node of the copied node.

        Returns:
            Node: Deep copy of the given node.
        """
        copy = Node(node.att_index, node.att_value)
        copy.parent = parent
        copy.left = Node.copy(
            node.left, parent=copy) if node.left is not None else None
        copy.right = Node.copy(
            node.right, parent=copy) if node.right is not None else None
        copy.fitness = node.fitness
        copy.y_true = node.y_true
        copy.y_pred = node.y_pred
        return copy

    def set_left(self, n):
        """
        Sets the left child of this node to the given node. Also sets the parent of the given node to this node.

        Args:
            n (Node): Node to be set as the left child.
        """
        self.left = n
        n.parent = self

    def set_right(self, n):
        """
        Sets the right child of this node to the given node. Also sets the parent of the given node to this node.

        Args:
            n (Node): Node to be set as the right child.
        """
        self.right = n
        n.parent = self

    def get_root(self):
        """
        Returns the root node of the tree.

        Returns:
            Node: Root node of the tree.
        """
        if self.parent is None:
            return self
        return self.parent.get_root()

    def get_children(self):
        """
        Returns the children of this node.

        Returns:
            tuple: Tuple containing the left and right children nodes.
        """
        return self.left, self.right

    def get_leaves(self):
        """
        Returns the leaves of this node.

        Returns:
            list: List of leaves of this node.
        """
        if self.left is None:
            return [self]
        return self.left.get_leaves() + self.right.get_leaves()

    def depth(self):
        """
        Returns the depth of this node.

        Returns:
            int: Depth of this node.
        """
        if self.parent is None:
            return 1
        return self.parent.depth() + 1

    def max_depth(self):
        """
        Returns the maximum depth of the tree. The depth of the tree is the number of edges on the longest path from the root to a leaf.

        Returns:
            int: Maximum depth of the tree.
        """
        return self.max_depth_helper(self.get_root())

    @staticmethod
    def max_depth_helper(n):
        """
        Helper function for max_depth.

        Args:
            n (Node): Node to be used as the root.

        Returns:
            int: Maximum depth of the tree.
        """
        if n is None:
            return 0
        l_depth = Node.max_depth_helper(n.left)
        r_depth = Node.max_depth_helper(n.right)
        return max(l_depth, r_depth) + 1

    def size(self):
        """
        Size is the number of all nodes (mid-tree nodes + leaves) in the trees.

        Returns:
            int: Number of all nodes in the trees.
        """
        return self.size_helper(self.get_root())

    @staticmethod
    def size_helper(n):
        """
        Helper function for size.

        Args:
            n (Node): Node to be used as the root.

        Returns:
            int: Number of all nodes in the trees.
        """
        if n is None:
            return 0
        return Node.size_helper(n.left) + Node.size_helper(n.right) + 1

    def make_node(self, depth=0, max_depth=None, random=None, att_indexes=None, att_values=None, class_count=None):
        """
        Randomly generates the node and its children. The depth of the tree and the maximum depth of the tree can be specified. The random number generator, attribute indexes, attribute values, and number of classes must be provided.

        Args:
            depth (int, optional): Current depth of the tree.
            max_depth (int, optional): Maximum depth of the tree.
            random (Random, optional): Random number generator.
            att_indexes (numpy.ndarray, optional): Attribute indexes.
            att_values (dict, optional): Attribute values.
            class_count (int, optional): Number of classes.

        Returns:
            Node: Randomly generated node with children.
        """
        node = None
        att_index = None
        value_index = None
        att_value = None
        if max_depth == None:
            max_depth = depth

        try:
            # if it's the root, first level or 50/50 chance of building new children.
            # Must be below maximal depth.
            if (depth <= 1 or (random.choice([True, False])) and depth < max_depth):
                subset_index = random.randint(0, len(att_indexes))
                att_index = att_indexes[subset_index]
                value_index = random.randint(0, len(att_values[att_index]))
                att_value = att_values[att_index][value_index]
                node = Node(att_index=att_index, att_value=att_value)
                node.left = self.make_node(depth=depth + 1, max_depth=max_depth, random=random,
                                           att_indexes=att_indexes, att_values=att_values, class_count=class_count)
                node.left.parent = node
                node.right = self.make_node(depth=depth + 1, max_depth=max_depth, random=random,
                                            att_indexes=att_indexes, att_values=att_values, class_count=class_count)
                node.right.parent = node
            else:  # result (leaf)
                r = random.randint(0, class_count)
                node = Node(att_index=-1, att_value=r)
        except Exception as e:
            print(f"{att_index};{att_value};{value_index}")
            print("Error:", e)
            node = None

        return node

    def clear_evaluation(self):
        """
        Clears the evaluation of this node and all its children. The true class values and predicted class values are reset to empty lists.
        """
        self.y_true = []
        self.y_pred = []
        if self.left:
            self.left.clear_evaluation()
        if self.right:
            self.right.clear_evaluation()

    def is_evaluated(self):
        """
        Returns true if this node and all its children are evaluated. A node is evaluated if it has true class values and predicted class values.

        Returns:
            bool: True if this node and all its children are evaluated.
        """
        if len(self.y_pred) == 0 and len(self.y_true) == 0:
            return False

        left_evaluated = self.left.is_evaluated() if self.left else True
        right_evaluated = self.right.is_evaluated() if self.right else True

        return left_evaluated or right_evaluated

    def predict_one(self, X, y=None, train=False):
        """
        Predicts the class of the given instance. If the actual class is provided, the true class value and predicted class value are stored for evaluation.

        Args:
            X (list): Instance to be predicted.
            y (int): Actual class of the given instance.
            train (bool): If it's used for training or only predicting

        Returns:
            int: Predicted class.
        """
        try:
            if self.att_index != -1:
                if X.iloc[self.att_index] > self.att_value:
                    if self.left is not None:
                        predicted = self.left.predict_one(X, y)
                else:
                    if self.right is not None:
                        predicted = self.right.predict_one(X, y)
            else:
                predicted = int(self.att_value)

            if train is True:
                if y is not None:
                    self.y_true.append(int(y))
                self.y_pred.append(predicted)

            return predicted
        except Exception as e:
            print(e)
            return -1

    def __str__(self):
        """
        Returns a string representation of this node.

        Returns:
            str: String representation of this node.
        """
        if self.att_index == -1:
            # Assuming att_value is a numeric index
            return str(int(self.att_value))
            # For nominal values, you might need a mapping from index to value
        else:
            return f"Attribute {self.att_index} > {self.att_value}"

    def __repr__(self):
        """
        Returns a string representation of this node.

        Returns:
            str: String representation of this node.
        """
        return self.__str__()



from gatree.tree.node import Node


class Mutation:
    """
    Class implementing the mutation operator for the genetic algorithm.
    """
    @staticmethod
    def mutation(root, att_indexes, att_values, class_count, random):
        """
        Mutation introduces random changes to a tree to maintain genetic diversity and explore new solutions. This helps in avoiding local optima by introducing new genetic structures.

        Pseudocode of the implementation:

        1. Randomly select a node within the tree.
        2. Depending on whether the selected node is a leaf or an internal node, apply different mutation strategies:

            - **Leaf Mutation:** Change the class label or replace the leaf with a new subtree.
            - **Internal Node Mutation:** Change the attribute, change the attribute value, replace the node with a class label, or replace the node with a new subtree.

        3. Ensure the mutated node maintains the structural integrity of the tree.
        4. Return the root of the mutated tree.

        Args:
            root (Node): The root node of the tree.
            att_indexes (list): List of attribute indexes.
            att_values (list): List of attribute values.
            class_count (int): Number of classes.
            random (Random): Random number generator.

        Returns:
            Node: The mutated node.
        """
        node = Node.copy(root)
        depth = node.max_depth()

        while True:
            if node.att_index == -1:  # for leaves
                Mutation.mutate_leaf(
                    node, att_indexes, att_values, class_count, random)
                break
            elif random.randint(0, depth) == 0:  # for mid-tree nodes
                Mutation.mutate_operator(
                    node, att_indexes, att_values, class_count, random)
                break

            # go to next random child
            if random.choice([True, False]):
                if node.left is not None:
                    node = node.left
                else:
                    break
            else:
                if node.right is not None:
                    node = node.right
                else:
                    break

        return node.get_root()

    @staticmethod
    def mutate_leaf(node, att_indexes, att_values, class_count, random):
        """
        Mutate a leaf node. The mutation operation can be one of the following:

        - Change the class of a leaf node.
        - Exchange a leaf node for a new subtree.

        Args:
            node (Node): The leaf node.
            att_indexes (list): List of attribute indexes.
            att_values (list): List of attribute values.
            class_count (int): Number of classes.
            random (Random): Random number generator.
        """
        if random.choice([True, False]):  # change class
            Mutation.change_class(node, class_count, random)
        else:  # exchange for new subtree
            Mutation.exchange_class_for_tree(
                node, att_indexes, att_values, class_count, random)

    @staticmethod
    def change_class(node, class_count, random):
        """
        Change class of a leaf node. The new class must be different from the old class.

        Args:
            node (Node): The leaf node.
            class_count (int): Number of classes.
            random (Random): Random number generator.
        """
        result_old = node.att_value
        result_new = result_old

        while result_old == result_new:  # classes must be different
            result_new = random.randint(0, class_count)

        node.att_value = result_new

    @staticmethod
    def exchange_class_for_tree(node, att_indexes, att_values, class_count, random):
        """
        Exchange a leaf node for a new subtree. The new subtree is created with a random depth.

        Args:
            node (Node): The leaf node.
            att_indexes (list): List of attribute indexes.
            att_values (list): List of attribute values.
            class_count (int): Number of classes.
            random (Random): Random number generator.
        """
        parent = node.parent
        left = False
        if parent.left == node:
            left = True

        n = Node()
        subtree = n.make_node(max_depth=node.depth(), random=random,
                              att_indexes=att_indexes, att_values=att_values, class_count=class_count)
        subtree.parent = parent

        if left:
            parent.set_left(subtree)
        else:
            parent.set_right(subtree)

    @staticmethod
    def mutate_operator(node, att_indexes, att_values, class_count, random):
        """
        Mutate a mid-tree node. The mutation operation can be one of the following:

        - Change the attribute of an operator node.
        - Change the attribute value of an operator node.
        - Exchange an operator node for a class.
        - Exchange an operator node for a new subtree.

        Args:
            node (Node): The mid-tree node.
            att_indexes (list): List of attribute indexes.
            att_values (list): List of attribute values.
            class_count (int): Number of classes.
            random (Random): Random number generator.
        """
        rand = random.random()
        if rand < 0.25 or node.parent == None:  # exchange attribute for other attribute
            Mutation.change_attribute(node, att_indexes, att_values, random)
        elif rand < 0.5:  # exchange attribute value for other value
            Mutation.change_attribute_value(node, att_values, random)
        elif rand < 0.75:  # exchange attribute for class
            Mutation.exchange_tree_for_class(node, class_count, random)
        else:  # exchange for new subtree
            Mutation.exchange_tree_for_tree(
                node, att_indexes, att_values, class_count, random)

    @staticmethod
    def change_attribute(node, att_indexes, att_values, random):
        """
        Change attribute of a mid-tree node. The new attribute must be different from the old attribute.

        Args:
            node (Node): The mid-tree node.
            att_indexes (list): List of attribute indexes.
            att_values (list): List of attribute values.
            random (Random): Random number generator.
        """
        att_index_old = node.att_index
        att_index_new = att_index_old

        att_value_old = node.att_value
        att_value_new = att_value_old

        while att_index_old == att_index_new:
            att_index_new = random.randint(0, len(att_indexes))
            att_value_new = random.randint(0, len(att_values[att_index_new]))

        node.att_index = att_index_new
        node.att_value = att_values[att_index_new][att_value_new]

    @staticmethod
    def change_attribute_value(node, att_values, random):
        """
        Change attribute value of a mid-tree node. The new attribute value must be different from the old attribute value.

        Args:
            node (Node): The mid-tree node.
            att_values (list): List of attribute values.
            random (Random): Random number generator.
        """
        att_index = node.att_index

        att_value_old = node.att_value
        att_value_new = att_value_old

        while att_value_old == att_value_new:
            att_value_new = random.randint(0, len(att_values[att_index]))

        node.att_value = att_values[att_index][att_value_new]

    @staticmethod
    def exchange_tree_for_class(node, class_count, random):
        """
        Exchange a mid-tree node for a class. The new class is selected randomly.

        Args:
            node (Node): The mid-tree node.
            class_count (int): Number of classes.
            random (Random): Random number generator.
        """
        parent = node.parent
        left = False
        if parent.left == node:
            left = True

        leaf = Node(att_index=-1, att_value=random.randint(0, class_count))
        leaf.parent = parent

        if left:
            parent.set_left(leaf)
        else:
            parent.set_right(leaf)

    @staticmethod
    def exchange_tree_for_tree(node, att_indexes, att_values, class_count, random):
        """
        Exchange a mid-tree node for a new subtree. The new subtree is created with a random depth.

        Args:
            node (Node): The mid-tree node.
            att_indexes (list): List of attribute indexes.
            att_values (list): List of attribute values.
            class_count (int): Number of classes.
            random (Random): Random number generator.
        """
        parent = node.parent
        left = False
        if parent.left == node:
            left = True

        n = Node()
        subtree = n.make_node(max_depth=node.depth(), random=random,
                              att_indexes=att_indexes, att_values=att_values, class_count=class_count)
        subtree.parent = parent

        if left:
            parent.set_left(subtree)
        else:
            parent.set_right(subtree)
